CARDINAL = "CARDINAL"
DATE = "DATE"
EVENT = "EVENT"
FAC = "FAC"
GPE = "GPE"
LANGUAGE = "LANGUAGE"
LAW = "LAW"
LOC = "LOC"
MONEY = "MONEY"
NORP = "NORP"
ORDINAL = "ORDINAL"
ORG = "ORG"
PERCENT = "PERCENT"
PERSON = "PERSON"
PRODUCT = "PRODUCT"
QUANTITY = "QUANTITY"
TIME = "TIME"
WORK_OF_ART = "WORK_OF_ART"

ALL_LABELS = [
    CARDINAL,
    DATE,
    EVENT,
    FAC,
    GPE,
    LANGUAGE,
    LAW,
    LOC,
    MONEY,
    NORP,
    ORDINAL,
    ORG,
    PERCENT,
    PERSON,
    PRODUCT,
    QUANTITY,
    TIME,
    WORK_OF_ART,
]
